// Copyright 2019 The Go Cloud Development Kit Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package docstore_test

import (
	"context"

	"gocloud.dev/internal/docstore"
)

func Example_actions() {
	ctx := context.Background()

	type Player struct {
		Name  string // Name is the key
		Score int
	}

	fred := &Player{Name: "Fred", Score: 18}
	got := &Player{Name: "Fred"}
	coll := docstore.NewCollection(nil)
	errs := coll.Actions().
		Put(fred).
		Get(got).
		Do(ctx)
	_ = errs
}
